
(* :Title: B-splines expressed analytically *)

(* :Authors: Robert Ipanaqu, Rubn T. Urbina & Segundo B. Correa *)

(* :Summary:
	  This package incorporates a few commands for obtaining and representing
	  some interesting curves in CAGD, as B-splines curves.
*)
	
(* :Context: CAGD`BsplinesCurves` *)

(* :Copyright:  Copyright 2003, Robert Ipanaqu & Rubn T. Urbina *)

(* :Package Version: 1.1 *)

(* :History:
	V1.0 by R. Ipanaqu, August 2003,
		developed at National University of Piura, PERU
	V1.1 by R. Ipanaque, S. Correa & R. Urbina, February 2004,
		developed at National University of Piura, PERU

*)

(* :Keywords: Uniform, Open, Bsplines, Curve *)

(* :Sources:
	Andres Iglesias and Flabio Gutierrez. Mathematica Package: "Curves.m",
		developed at Cantabria University, SPAIN (1997).
	David F. Rogers and J. Alan Adams. "Mathematical elements for
		computer graphics" (1990).
	Ruben Urbina y Robert Ipanaque. Thesis of Degree (1999).
*)

(* :Mathematica Version: 5.1 *)

(* :Limitations: No limitations *)


            (* Start of the package *)

BeginPackage["BsplinesCurves`"]


              (* Usage messages *)

Uniform::usage = "Uniform[nmas1, k] is a command that returns an uniform
	knots vector of order 'k', for 'nmas1' points. Remember that
	2 <= k <= nmas1."

Open::usage = "Open[nmas1, k] is a command that returns an open	knots vector
	of order 'k', for 'nmas1' points. Remember that	2 <= k <= nmas1."

Bsplines::usage = "Bsplines[(var,) k, x] is a command which admits a
	variable 'var', order 'k' and knots sequence 'x' and returns the
	B-splines function of order 'k' defined regarding the knots sequence
	'x'. In the case of not entering the variable is assumed that it is 't'."

Curve::usage = "Curve[(var,) pts (, ops)] is a command which admits a 
        set of points ('pts') and a variable ('var') and returns the B-splines
	curve defined by this set of points. This command incorporates the
	options: {Knots -> Open, Order -> 3}. In the case of not entering the
	variable is assumed that it is 't'."

PiecewiseForm::usage =
	"PiecewiseForm[f] is a command that allows to change the presentation
	of the function f.
	PiecewiseForm[{c}] is a command that allows to change the presentation
	of the curve c."


              (* Error messages *)

Curve::badorder = "Order must be no greater than ``; received ``."

Knots::badorder = "Order must be no greater than ``; received ``."

Knots::badsequence = "The sequence `` should be monotonous and non falling."

Knots::badlength = "Expecting `` knots but received ``."

Points::baddimensions = "The components of `` don't have the same dimensions."

t::usage = "t."


            (* Options definitions *)

Options[Curve] = {Knots -> Open, Order -> 3}


        (* Start of the private context *)

Begin["`Private`"]

Uniform[nmas1_Integer /; nmas1 > 1, k_Integer /; k > 0] :=
  Module[{},

    If[ Not[k <= nmas1],
	Return[Message[Knots::badorder, nmas1, k]] ];

    Table[i - 1, {i, nmas1 + k}]
    ]

Unprotect[Open]

Open[nmas1_Integer /; nmas1 > 1, k_Integer /; k > 0] :=
  Module[{},

    If[ Not[k <= nmas1],
	Return[Message[Knots::badorder, nmas1, k]] ];

    Table[
      Which[i <= k, 0, i <= nmas1, i - k, True, nmas1 - 1 - k + 2],
          {i, nmas1 + k}]
    ]

Protect[Open]

q = If[#2 === 0, 0, #1/#2] &;

B = If[#3 == 1,
      If[#5 === #2, 1, 0],
      q[#1 - #4[[#2]], #4[[#2 + #3 - 1]] - #4[[#2]]]*
          B[#1, #2, #3 - 1, #4, #5] +
        q[#4[[#2 + #3]] - #1, #4[[#2 + #3]] - #4[[#2 + 1]]]*
          B[#1, #2 + 1, #3 - 1, #4, #5]] &

a = {#1 < #2[[1]], Sequence @@ DeleteCases[
	Array[If[#2[[j]] =!= #2[[j + 1]], #1 < #2[[j + 1]]] &, #3], 
					Null], #1 >= #2[[#3 + 1]]} &

b = DeleteCases[
	Array[If[#2[[j]] =!= #2[[j + 1]], Together[#4[[j]]]] &, #3],
					Null] &

w = Piecewise[Thread[{#2, #1}]] &

Bspline[var_:t, k_Integer /; k > 0, Ops___Rule] :=
  Module[{x, f},

    x = Knots /. Flatten[{Ops}] /. Options[Bspline];

    If[ x === Default, x = Range[0, k] ];
    If[ Head[Plus @@ x] == List, Return[Message[Knots::badhead, x]]  ];
    If[ Length[x] != k + 1,
	Return[Message[Knots::badlength, k + 1, Length[x]]] ];
    If[ Not[OrderedQ[N[x]]], Return[Message[Knots::badsequence, x]] ];

    f = Array[B[var, 1, k, x, #] &, k];

    w[a[var, x, k], Insert[b[var, x, k, f], 0, {{-1}, {1}}]]
    ]

Curve[var_:t, pp_List, Ops___Rule] :=
  Module[{nmas1 = Length[pp], aux = Union[Map[Dimensions, pp]],
		F, c, r, k, x, p},

    {k, x, r} = {Order, Knots, Range} /. Flatten[{Ops}] /. Options[Curve];

    If[ x === Uniform, x = Uniform[nmas1, k] ];
    If[ x === Open, x = Open[nmas1, k] ];
    If[ 0 + Sequence @@ Flatten[aux] == 0, p = Map[List, pp], p = pp ];
    If[ Not[2 <= k <= nmas1],
	Return[Message[Curve::badorder, nmas1, k]] ];
    If[ Length[x] != nmas1 + k,
	Return[Message[Knots::badlength, nmas1 + k, Length[x]]] ];
    If[ Not[OrderedQ[N[x]]], Return[Message[Knots::badsequence, x]] ];
    If[ Length[aux] != 1, Return[Message[Points::baddimensions, p]] ];

    If[Head[k] =!= Integer || Head[x] =!= List,

       SequenceForm[ "Curve[",var,",",p,",<>]" ],

       F = Array[Sum[p[[i]] * B[var, i, k, x, #], {i, nmas1}] &, nmas1 + k - 1];
       c = w[a[var, x, nmas1 + k - 1],
                      Insert[b[var, x, nmas1 + k - 1, F],
				Table[0, {Length[p[[1]]]}], {{-1}, {1}}]];

    If[r, Sequence @@ {c, {var, x[[k]], x[[nmas1 + 1]] - 10.^(-8)}}, c]
      ]
    ]


         (* End of the private context *)

End[]


             (* End of the package *)

EndPackage[]
